@@ -40,6 +40,7 @@
 
 #include <pdal/ArtifactManager.hpp>
 #include <pdal/Polygon.hpp>
+#include <pdal/private/OGRSpec.hpp>
 #include <pdal/SrsBounds.hpp>
 #include <pdal/pdal_features.hpp>
 #include <pdal/util/ThreadPool.hpp>
@@ -61,17 +62,17 @@ const StaticPluginInfo s_info
 {
     "readers.ept",
     "EPT Reader",
-    "http://pdal.io/stages/reader.ept.html",
+    "https://pdal.org/stages/reader.ept.html",
     { "ept" }
 };
 
-void reprogrow(BOX3D& b, SrsTransform& xform, double x, double y, double z)
+void reprogrow(BOX3D& b, const SrsTransform& xform, double x, double y, double z)
 {
     xform.transform(x, y, z);
     b.grow(x, y, z);
 }
 
-BOX3D reprojectBoundsViaCorner(BOX3D src, SrsTransform& xform)
+BOX3D reprojectBoundsViaCorner(BOX3D src, const SrsTransform& xform)
 {
     if (!xform.valid())
         return src;
@@ -90,7 +91,7 @@ BOX3D reprojectBoundsViaCorner(BOX3D src, SrsTransform& xform)
     return b;
 }
 
-BOX3D reprojectBoundsBcbfToLonLat(BOX3D src, SrsTransform& xform)
+BOX3D reprojectBoundsBcbfToLonLat(BOX3D src, const SrsTransform& xform)
 {
     if (!xform.valid())
         return src;
@@ -107,7 +108,7 @@ BOX3D reprojectBoundsBcbfToLonLat(BOX3D src, SrsTransform& xform)
     }
 
     // Round the minimum longitude up to the nearest multiple of 90 degrees.
-    int x = std::ceil(src.minx);
+    int x = (int) std::ceil(src.minx);
     const int remainder = std::abs(x) % 90;
     if (x < 0)
         x = -(std::abs(x) - remainder);
@@ -157,10 +158,7 @@ public:
     double m_resolution = 0;
     std::vector<Polygon> m_polys;
     NL::json m_addons;
-
-    NL::json m_query;
-    NL::json m_headers;
-    NL::json m_ogr;
+    OGRSpec m_ogr;
     bool m_ignoreUnreadable = false;
 };
 
@@ -174,11 +172,21 @@ public:
     std::unique_ptr<ept::Hierarchy> hierarchy;
     std::queue<ept::TileContents> contents;
     ept::AddonList addons;
-    std::mutex mutex;
+    mutable std::mutex mutex;
     std::condition_variable contentsCv;
     std::vector<PolyXform> polys;
     BoxXform bounds;
     SrsTransform llToBcbfTransform;
+    uint64_t depthEnd {0};    // Zero indicates selection of all depths.
+    uint64_t hierarchyStep {0};
+    uint64_t nodeId;
+
+    void overlaps(ept::Hierarchy& target, const NL::json& hier, const ept::Key& key);
+    bool passesSpatialFilter(const BOX3D& tileBounds) const;
+    bool hasSpatialFilter() const
+    {
+        return !polys.empty() || bounds.box.valid();
+    }
 };
 
 EptReader::EptReader() : m_args(new EptReader::Args), m_p(new EptReader::Private),
@@ -201,37 +209,12 @@ void EptReader::addArgs(ProgramArgs& args)
     args.add("polygon", "Bounding polygon(s) to crop requests",
         m_args->m_polys).setErrorText("Invalid polygon specification. "
             "Must be valid GeoJSON/WKT");
-    args.add("header", "Header fields to forward with HTTP requests", m_args->m_headers);
-    args.add("query", "Query parameters to forward with HTTP requests", m_args->m_query);
     args.add("ogr", "OGR filter geometries", m_args->m_ogr);
     args.add("ignore_unreadable", "Ignore errors for missing point data nodes",
         m_args->m_ignoreUnreadable);
 }
 
 
-void EptReader::setForwards(StringMap& headers, StringMap& query)
-{
-    try
-    {
-        if (!m_args->m_headers.is_null())
-            headers = m_args->m_headers.get<StringMap>();
-    }
-    catch (const std::exception& err)
-    {
-        throwError(std::string("Error parsing 'headers': ") + err.what());
-    }
-
-    try
-    {
-        if (!m_args->m_query.is_null())
-            query = m_args->m_query.get<StringMap>();
-    }
-    catch (const std::exception& err)
-    {
-        throwError(std::string("Error parsing 'query': ") + err.what());
-    }
-}
-
 void EptReader::initialize()
 {
     auto& debug(log()->get(LogLevel::Debug));
@@ -242,10 +225,7 @@ void EptReader::initialize()
             threads << " threads" << std::endl;
     m_p->pool.reset(new ThreadPool(threads));
 
-    StringMap headers;
-    StringMap query;
-    setForwards(headers, query);
-    m_p->connector.reset(new connector::Connector(headers, query));
+    m_p->connector.reset(new connector::Connector(m_filespec));
 
     try
     {
@@ -258,12 +238,8 @@ void EptReader::initialize()
         throwError(err.what());
     }
 
-    if (!m_args->m_ogr.is_null())
-    {
-        auto& plist = m_args->m_polys;
-        std::vector<Polygon> ogrPolys = gdal::getPolygons(m_args->m_ogr);
-        plist.insert(plist.end(), ogrPolys.begin(), ogrPolys.end());
-    }
+    std::vector<Polygon> ogrPolys = m_args->m_ogr.getPolygons();
+    m_args->m_polys.insert(m_args->m_polys.end(), ogrPolys.begin(), ogrPolys.end());
 
     // Create transformations from our source data to the bounds SRS.
     if (m_args->m_bounds.valid())
@@ -323,7 +299,7 @@ void EptReader::initialize()
     // Figure out our max depth.
     const double queryResolution(m_args->m_resolution);
     //reseting depthEnd if initialize() has been called before
-    m_depthEnd = 0;
+    m_p->depthEnd = 0;
     if (queryResolution)
     {
         double currentResolution =
@@ -333,17 +309,17 @@ void EptReader::initialize()
 
         // To select the current resolution level, we need depthEnd to be one
         // beyond it - this is a non-inclusive parameter.
-        ++m_depthEnd;
+        ++m_p->depthEnd;
 
         while (currentResolution > queryResolution)
         {
             currentResolution /= 2;
-            ++m_depthEnd;
+            ++m_p->depthEnd;
         }
 
         debug << "Query resolution:  " << queryResolution << "\n";
         debug << "Actual resolution: " << currentResolution << "\n";
-        debug << "Depth end: " << m_depthEnd << "\n";
+        debug << "Depth end: " << m_p->depthEnd << "\n";
     }
 
     debug << "Query bounds: " << m_p->bounds.box << "\n";
@@ -456,7 +432,7 @@ void EptReader::handleOriginQuery()
     {
         BOX3D q(toBox3d(found.at("bounds")));
         // Bloat the query bounds slightly (we'll choose one tick of the EPT's
-        // scale) to make sure we don't miss any points due to them being 
+        // scale) to make sure we don't miss any points due to them being
         // precisely on the bounds edge.
         q.grow(
             (std::max)(
@@ -495,7 +471,7 @@ QuickInfo EptReader::inspect()
     // If there is a spatial filter from an explicit --bounds, an origin query,
     // or polygons, then we'll limit our number of points to be an upper bound,
     // and clip our bounds to the selected region.
-    if (hasSpatialFilter())
+    if (m_p->hasSpatialFilter())
     {
         log()->get(LogLevel::Debug) <<
             "Determining overlapping point count" << std::endl;
@@ -553,6 +529,8 @@ void EptReader::addDimensions(PointLayoutPtr layout)
 // stick the tile on the queue and notify the main thread.
 void EptReader::load(const ept::Overlap& overlap)
 {
+    using namespace std::chrono_literals;
+
     m_p->pool->add([this, overlap]()
         {
             // Read the tile.
@@ -571,9 +549,23 @@ void EptReader::load(const ept::Overlap& overlap)
                 // Put the tile on the output queue.  Note that if the tile has
                 // an error and ignoreUnreadable isn't set, this will be fatal
                 // but that will occur downstream outside of this pool thread.
-                std::unique_lock<std::mutex> l(m_p->mutex);
-                m_p->contents.push(std::move(tile));
-                l.unlock();
+
+                // Loop to push the tile on the queue until there is room.
+                while (true)
+                {
+                    {
+                        std::lock_guard<std::mutex> l(m_p->mutex);
+                        if (m_p->contents.size() < m_p->pool->numThreads())
+                        {
+                            m_p->contents.push(std::move(tile));
+                            break;
+                        }
+                    }
+                    // No room on queue, sleep. Could do a condition variable but that's
+                    // more complex and probably makes no difference in most cases where
+                    // this would come up.
+                    std::this_thread::sleep_for(50ms);
+                }
             }
             else
             {
@@ -597,17 +589,17 @@ void EptReader::ready(PointTableRef table)
     m_pointIdDim = table.layout()->findDim("EptPointId");
 
     if (
-        m_queryOriginId != -1 && 
+        m_queryOriginId != -1 &&
         !table.layout()->hasDim(Dimension::Id::OriginId))
     {
         // In this case we can't compare the OriginId for each point since the
-        // EPT data does not have that attribute saved.  We will keep the 
+        // EPT data does not have that attribute saved.  We will keep the
         // spatial query to limit the data to the extents of the requested
         // origin, but if other origins overlap these extents, then their points
         // will also be included.
         m_queryOriginId = -1;
 
-        log()->get(LogLevel::Warning) << 
+        log()->get(LogLevel::Warning) <<
             "An origin query was given but no OriginId dimension exists - " <<
             "points from other origins may be included" << std::endl;
     }
@@ -631,7 +623,7 @@ void EptReader::ready(PointTableRef table)
     if (overlapPoints > 1e8)
     {
         log()->get(LogLevel::Warning) << overlapPoints <<
-            " will be downloaded" << std::endl;
+            " points will be downloaded" << std::endl;
     }
 
     m_pointId = 0;
@@ -661,58 +653,50 @@ void EptReader::overlaps()
     key.b = m_p->info->bounds();
 
     {
-        m_nodeId = 1;
+        m_p->nodeId = 1;
         std::string filename = m_p->info->hierarchyDir() + key.toString() + ".json";
 
         // First, determine the overlapping nodes from the EPT resource.
-        overlaps(*m_p->hierarchy, m_p->connector->getJson(filename), key);
+        m_p->overlaps(*m_p->hierarchy, m_p->connector->getJson(filename), key);
     }
     m_p->pool->await();
 
     // Determine the addons that exist to correspond to tiles.
     for (auto& addon : m_p->addons)
     {
-        m_nodeId = 1;
+        m_p->nodeId = 1;
         std::string filename = addon.hierarchyDir() + key.toString() + ".json";
-        overlaps(addon.hierarchy(), m_p->connector->getJson(filename), key);
+        m_p->overlaps(addon.hierarchy(), m_p->connector->getJson(filename), key);
         m_p->pool->await();
     }
 }
 
 
-bool EptReader::hasSpatialFilter() const
-{
-    return !m_p->polys.empty() || m_p->bounds.box.valid();
-}
-
-
 // Determine if an EPT tile overlaps our query boundary
-bool EptReader::passesSpatialFilter(const BOX3D& tileBounds) const
+bool EptReader::Private::passesSpatialFilter(const BOX3D& tileBounds) const
 {
     auto boxOverlaps = [this, &tileBounds]() -> bool
     {
-        if (!m_p->bounds.box.valid())
+        if (!bounds.box.valid())
             return true;
 
-        if (m_p->llToBcbfTransform.valid())
+        if (llToBcbfTransform.valid())
         {
-            return reprojectBoundsBcbfToLonLat(m_p->bounds.box, m_p->llToBcbfTransform)
-                .overlaps(tileBounds);
+            return reprojectBoundsBcbfToLonLat(bounds.box, llToBcbfTransform).overlaps(tileBounds);
         }
 
         // If the reprojected source bounds doesn't overlap our query bounds, we're done.
-        return reprojectBoundsViaCorner(tileBounds, m_p->bounds.xform)
-            .overlaps(m_p->bounds.box);
+        return reprojectBoundsViaCorner(tileBounds, bounds.xform).overlaps(bounds.box);
     };
 
     // Check the box of the key against our query polygon(s). If it doesn't overlap,
     // we can skip
     auto polysOverlap = [this, &tileBounds]() -> bool
     {
-        if (m_p->polys.empty())
+        if (polys.empty())
             return true;
 
-        for (auto& ps : m_p->polys)
+        for (auto& ps : polys)
             if (!ps.poly.disjoint(reprojectBoundsViaCorner(tileBounds, ps.xform)))
                 return true;
         return false;
@@ -725,12 +709,12 @@ bool EptReader::passesSpatialFilter(const BOX3D& tileBounds) const
     // This lock is here because if a bunch of threads are using the transform
     // at the same time, it seems to get corrupted. There may be other instances
     // that need to be locked.
-    std::lock_guard<std::mutex> lock(m_p->mutex);
+    std::lock_guard<std::mutex> lock(mutex);
     return boxOverlaps() && polysOverlap();
 }
 
 
-void EptReader::overlaps(ept::Hierarchy& target, const NL::json& hier, const ept::Key& key)
+void EptReader::Private::overlaps(ept::Hierarchy& target, const NL::json& hier, const ept::Key& key)
 {
     // If our key isn't in the hierarchy, we've totally traversed this tree
     // branch (there are no lower nodes).
@@ -740,7 +724,7 @@ void EptReader::overlaps(ept::Hierarchy& target, const NL::json& hier, const ept
 
     // If our query geometry doesn't overlap the tile or we're past the end of the requested
     // depth, return.
-    if (!passesSpatialFilter(key.b) || (m_depthEnd && key.d >= m_depthEnd))
+    if (!passesSpatialFilter(key.b) || (depthEnd && key.d >= depthEnd))
         return;
 
 
@@ -754,28 +738,28 @@ void EptReader::overlaps(ept::Hierarchy& target, const NL::json& hier, const ept
 
     if (numPoints == -1)
     {
-        if (!m_hierarchyStep)
-            m_hierarchyStep = key.d;
+        if (!hierarchyStep)
+            hierarchyStep = key.d;
 
         // If the hierarchy points value here is -1, then we need to fetch the
         // hierarchy subtree corresponding to this root.
-        m_p->pool->add([this, &target, key]()
+        pool->add([this, &target, key]()
         {
             try
             {
-                std::string filename = m_p->info->hierarchyDir() + key.toString() + ".json";
-                const auto subRoot(m_p->connector->getJson(filename));
+                std::string filename = info->hierarchyDir() + key.toString() + ".json";
+                const auto subRoot(connector->getJson(filename));
                 overlaps(target, subRoot, key);
             }
             catch (const arbiter::ArbiterError& err)
             {
-                throwError(err.what());
+                throw pdal_error(err.what());
             }
         });
     }
     else if (numPoints < 0)
     {
-        throwError("Invalid point count for key '" + key.toString() + "'.");
+        throw pdal_error("Invalid point count for key '" + key.toString() + "'.");
     }
     else
     {
@@ -783,8 +767,8 @@ void EptReader::overlaps(ept::Hierarchy& target, const NL::json& hier, const ept
         // not match the base hierarchy, but it doesn't matter since
         // they are never used.
         {
-            std::lock_guard<std::mutex> lock(m_p->mutex);
-            target.emplace(key, (point_count_t)numPoints, m_nodeId++);
+            std::lock_guard<std::mutex> lock(mutex);
+            target.emplace(key, (point_count_t)numPoints, nodeId++);
         }
 
         for (uint64_t dir(0); dir < 8; ++dir)
@@ -818,7 +802,7 @@ bool EptReader::processPoint(PointRef& dst, const ept::TileContents& tile)
     PointId pointId = m_pointId++;
 
     PointRef p(t, pointId);
-    if (m_queryOriginId != -1 && 
+    if (m_queryOriginId != -1 &&
         p.getFieldAs<int64_t>(Id::OriginId)!= m_queryOriginId)
     {
         return false;
@@ -855,7 +839,7 @@ bool EptReader::processPoint(PointRef& dst, const ept::TileContents& tile)
     double z = p.getFieldAs<double>(Id::Z);
 
     // If there is a spatial filter, make sure it passes.
-    if (hasSpatialFilter())
+    if (m_p->hasSpatialFilter())
         if (!passesBoundsFilter(x, y, z) || !passesPolyFilter(x, y, z))
             return false;
 
@@ -934,7 +918,7 @@ point_count_t EptReader::read(PointViewPtr view, point_count_t count)
     {
         ept::ArtifactPtr artifact
             (new ept::Artifact(std::move(m_p->info), std::move(m_p->hierarchy),
-                std::move(m_p->connector), m_hierarchyStep));
+                std::move(m_p->connector), m_p->hierarchyStep));
         m_artifactMgr->put("ept", artifact);
     }
 
